package tofly.opc_gateway;

import javafish.clients.opc.JOpc;
import javafish.clients.opc.component.OpcGroup;
import javafish.clients.opc.component.OpcItem;
import javafish.clients.opc.exception.*;
import javafish.clients.opc.lang.Translate;
import javafish.clients.opc.variant.Variant;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.scheduling.TaskScheduler;
import org.springframework.scheduling.support.PeriodicTrigger;
import org.springframework.stereotype.Component;
import tofly.opc_gateway.config.OpcServerProperties;
import tofly.opc_gateway.log.Log;
import tofly.opc_gateway.log.LogFactory;
import tofly.opc_gateway.service.StoreService;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;

@Component
public class DataAcquisitionRunner implements ApplicationRunner {


    private static final Log log = LogFactory.getLog("DataAcquisition");

    @Value("${spring.datasource.platform}")
    private String platform;

    @Autowired
    private OpcServerProperties opcServerProperties;
    @Autowired
    private TaskScheduler taskScheduler;
    @Autowired
    private StoreService storeService;

    public DataAcquisitionRunner() {
        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            JOpc.coUninitialize();
            // 在JVM关闭之前执行收尾工作
            // 注意事项：
            // 1.在这里执行的动作不能耗时太久
            // 2.不能在这里再执行注册，移除关闭钩子的操作
            // 3 不能在这里调用System.exit()
            // log.info("do shutdown hook");
        }));
    }

    @Override
    public void run(ApplicationArguments args) throws Exception {
        JOpc.coInitialize();
        opcServerProperties.getMappings().forEach((k, v) -> {
            try {
                MDC.put("opc_server", k);
                String host = v.getHost();
                String progId = v.getProgId();
                Map<String, OpcServerProperties.DeviceConfiguration> deviceMappings = v.getDeviceMappings();

                JOpc jopc = new JOpc(host, progId, "JOPC1");
                List<OpcGroup> groups = new ArrayList<>();
                Map<String, List<String>> varMap = new HashMap<>();
                deviceMappings.forEach((dk, dv) -> {
                    OpcGroup group = new OpcGroup(dk, true, 500, 0.0f);
                    Map<String, OpcServerProperties.FieldConfiguration> fieldMappings = dv.getFieldMappings();

                    fieldMappings.forEach((fk, fv) -> {
                        String[] vars = fk.split(",");
                        Arrays.asList(vars).forEach(var -> {
                            if (!varMap.containsKey(var)) {
                                group.addItem(new OpcItem(var, true, ""));
                                List<String> fieldKeys = new ArrayList<>();
                                fieldKeys.add(fk);
                                varMap.put(var, fieldKeys);
                            } else {
                                List<String> fieldKeys = varMap.get(var);
                                fieldKeys.add(fk);
                                varMap.put(var, fieldKeys);
                            }
                        });
                    });

                    groups.add(group);
                });

                groups.forEach(group -> {
                    jopc.addGroup(group);
                });
                AtomicBoolean disconnected = new AtomicBoolean(false);
                try {
                    conn(jopc);
                } catch (Exception e) {
                    log.error("连接失败", e);
                    disconnected.set(true);
                }
                PeriodicTrigger trigger = new PeriodicTrigger(v.getPeriod());
                trigger.setInitialDelay(1000);

                taskScheduler.schedule(() -> {
                    MDC.put("opc_server", k);
                    try {
                        doAcquisition(jopc, groups, varMap, deviceMappings, disconnected);
                    } finally {
                        MDC.clear();
                    }
                }, trigger);

            } catch (Throwable e) {
                log.error("采集数据异常", e);
            } finally {
                MDC.clear();
            }
        });
    }

    private void doAcquisition(JOpc jopc,
                               List<OpcGroup> groups,
                               Map<String, List<String>> varMap,
                               Map<String, OpcServerProperties.DeviceConfiguration> deviceMappings,
                               AtomicBoolean disconnected) {
        clearConsole();
        if (disconnected.get()) {
            try {
                log.info(String.format("%s正在重连", jopc.getServerProgID()));
                conn(jopc);
                disconnected.set(false);
            } catch (ConnectivityException | UnableAddGroupException | UnableAddItemException e) {
                log.error(String.format("%s重连失败", jopc.getServerProgID()), e);
                try {
                    Thread.sleep(5000);
                } catch (InterruptedException e2) {
                    // ignore...
                }
                return;
            }
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                // ignore...
            }
        }
        groups.forEach(group -> {

            ArrayList<OpcItem> items = group.getItems();
            String groupName = group.getGroupName();
            String deviceId = groupName;
            OpcServerProperties.DeviceConfiguration deviceConfiguration = deviceMappings.get(deviceId);
            String table = deviceConfiguration.getTable();
            Map<String, OpcServerProperties.FieldConfiguration> fieldMappings = deviceConfiguration.getFieldMappings();
            Map<String, BigDecimal> valueMap = new LinkedHashMap<>();
            items.forEach(item -> {
                try {
                    OpcItem responseItem = jopc.synchReadItem(group, item);
                    String itemName = item.getItemName();
                    List<String> fieldKeys = varMap.get(itemName);
                    for (int j = 0; j < fieldKeys.size(); j++) {
                        String fieldKey = fieldKeys.get(j);
                        OpcServerProperties.FieldConfiguration fieldConfiguration = fieldMappings.get(fieldKey);
                        boolean isIntType = fieldConfiguration.isIntType();
                        double magnification = fieldConfiguration.getMagnification();
                        Integer digits = fieldConfiguration.getDigits();
                        double cardinality = fieldConfiguration.getCardinality();
                        Object value1 = getValue(responseItem.getValue(), isIntType);

                        BigDecimal x = new BigDecimal(String.valueOf(value1));
                        if (magnification != 1D) {
                            x = x.multiply(new BigDecimal(String.valueOf(magnification)));
                        }
                        BigDecimal value = null;
                        if (fieldKey.contains(",")) {
                            BigDecimal currentValue = valueMap.getOrDefault(fieldKey, new BigDecimal(cardinality));
                            value = currentValue.add(x);
                        } else {
                            value = new BigDecimal(cardinality).add(x);
                        }

                        if (digits != null) {
                            value = value.setScale(digits, RoundingMode.HALF_UP);
                        }
                        valueMap.put(fieldKey, value.stripTrailingZeros());
                    }


                } catch (SynchReadException e) {
                    disconnected.set(true);
                    log.error("采集数据异常", e);
                    try {
                        jopc.unregisterGroups();
                    } catch (UnableRemoveGroupException ex) {
                        //ignore...
                    }
                }
            });
            if (valueMap.isEmpty()) return;

            Map<String, String> newValueMap = new LinkedHashMap<>();
            fieldMappings.forEach((fk, fv) -> {
                newValueMap.put(fk, valueMap.get(fk).toPlainString());
            });
            Map<String, Object> dataMap = new LinkedHashMap<>();
            newValueMap.forEach((vk, vv) -> {
                OpcServerProperties.FieldConfiguration fieldConfiguration = fieldMappings.get(vk);
                String variantName = fieldConfiguration.getName();
                String label = fieldConfiguration.getLabel();
                String value = vv;
                log.info(label == null ? "[UNDEFINED]" : label + ":" + value);
                dataMap.put(variantName, value);
            });
            storeService.store(table, deviceId, dataMap);
        });
    }


    private void clearConsole() {
        try {
            String operatingSystem = System.getProperty("os.name"); //Check the current operating system
            if (operatingSystem.contains("Windows")) {
                ProcessBuilder pb = new ProcessBuilder("cmd", "/c", "cls");
                Process startProcess = pb.inheritIO().start();
                startProcess.waitFor();
            } else {
                ProcessBuilder pb = new ProcessBuilder("clear");
                Process startProcess = pb.inheritIO().start();
                startProcess.waitFor();
            }
        } catch (Exception e) {
            log.error("clearConsole error", e);
        }
    }

    private static void conn(JOpc jopc) throws ConnectivityException, UnableAddGroupException, UnableAddItemException {
        jopc.connect();
        log.info(String.format("%s正在连接", jopc.getServerProgID()));
        jopc.registerGroups();
        log.info(String.format("%s连接成功", jopc.getServerProgID()));
    }

    private Object getValue(Variant value, boolean isIntType) {
        int variantType = value.getVariantType();
        //String variantName = Variant.getVariantName(variantType);
        //log.info(variantType+":"+variantName);
        switch (variantType) {
            case 0:
            case 1:
                return "";
            case 2:
                return value.getWord();
            case 4:
                if (isIntType) {
                    return new Float(value.getFloat()).intValue();
                } else {
                    return value.getFloat();
                }
            case 5:
                if (isIntType) {
                    return new Double(value.getDouble()).intValue();
                } else {
                    return value.getDouble();
                }
            case 6:
            case 7:
            case 8:
            case 30:
            case 31:
                return value.getString();
            case 10:
                return "ERROR";
            case 11:
                return value.getBoolean() ? 1 : 0;
            case 16:
            case 17:
                return value.getByte();
            case 22:
                return value.getInteger();
            default:
                throw new VariantTypeException(Translate.getString("VARIANT_TYPE_EXCEPTION"));
        }
    }
}
