package tofly.opc_gateway.service;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.CannotGetJdbcConnectionException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
@Slf4j
public class StoreServiceImpl implements StoreService {
    @Autowired
    private JdbcTemplate jdbcTemplate;

    private static final String storeDir = "failure";
    private static final String suffix = ".properties";

    @Override
    @Async
    public void store(String table, String deviceId, Map<String, Object> data) {
        Date now = new Date();
        try {
            store0(table, now, data);
        } catch (Exception e) {
            log.warn(String.format("保存数据失败:%s,%s", deviceId, data), e);
            storeLocalFile(table, now.getTime(), deviceId, data);
        }
    }

    private void store0(String table, Date time, Map<String, Object> data) {
        Set<String> keys = data.keySet();
        for (String vk : keys) {
            Object vv = data.get(vk);
            String sql = String.format("SELECT * FROM %s WHERE code = ? and scada_time = ?", table);
            List<Map<String, Object>> maps = jdbcTemplate.queryForList(sql, vk, time);
            if (maps.size() > 0) {
                update(table, time, vk, String.valueOf(vv));
            } else {
                save(table, time, vk, String.valueOf(vv));
            }
        }
    }

    @Override
    public int checkAndRestore() {
        Path path = Paths.get(storeDir);
        if (!Files.exists(path)) return 0;
        int success = 0;
        try (Stream<Path> pathStream = Files.walk(path, 1)) {
            List<Path> paths = pathStream.filter(e -> e.toFile().isFile()).collect(Collectors.toList());
            for (Path file : paths) {
                String name = file.toFile().getName();
                int i = name.lastIndexOf(suffix);
                String str = name;
                if (i > 0) {
                    str = name.substring(0, i);
                }
                String[] s = str.split("_");
                if (s.length < 3) continue;
                Properties properties = new Properties();
                FileReader reader = null;
                try {
                    reader = new FileReader(file.toFile());
                    properties.load(reader);
                } catch (IOException ex) {
                    log.warn("加载文件异常", ex);
                    continue;
                } finally {
                    if (reader != null) {
                        try {
                            reader.close();
                        } catch (IOException ex) {
                            //ignore...
                        }
                    }
                }

                Set<String> propertyNames = properties.stringPropertyNames();
                Map<String, Object> data = new LinkedHashMap<>();
                propertyNames.forEach(p -> data.put(p, properties.get(p)));
                try {
                    Date time = new Date();
                    time.setTime(Long.valueOf(s[1]));
                    String table = s[2];
                    store0(table, time, data);
                    Files.delete(file);
                    success++;
                    if (success % 10 == 0) {
                        log.info(String.format("异常检测恢复%s条", success));
                    }
                } catch (Exception ex) {
                    if (ex instanceof CannotGetJdbcConnectionException) {
                        log.info(String.format("异常检测恢复失败(退出检测):%s，原因:%s", name, ex.getMessage()));
                        break;
                    } else {
                        log.info(String.format("异常检测恢复失败:%s，原因:%s", name, ex.getMessage()));
                    }
                }
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return success;
    }

    private void storeLocalFile(String table, long time, String deviceId, Map<String, Object> data) {
        Properties properties = new Properties();
        data.forEach((k, v) -> properties.put(k, String.valueOf(v)));
        Path temp = Paths.get(storeDir);
        if (!Files.exists(temp)) {
            try {
                Files.createDirectory(temp);
            } catch (IOException ex) {
                log.warn("创建目录异常", ex);
            }
        }
        FileWriter writer = null;
        try {
            Path path =  Paths.get(storeDir, String.format("%s_%s_%s%s", deviceId, time, table, suffix));
            if (!Files.exists(path)) {
                path = Files.createFile(path);
            }
            writer = new FileWriter(path.toFile());
            properties.store(writer, deviceId);
            writer.flush();
            writer.close();
        } catch (IOException ex) {
            log.warn("存储文件异常", ex);
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    //ignore...
                }
            }
        }

    }

    private void save(String table, Date time, String code, String value) {
        String sql = String.format("INSERT INTO %s (code, scada_time, value) VALUES (?, ?, ?)", table);
        jdbcTemplate.update(sql, code, time, value);
    }

    private void update(String table, Date time, String code, String value) {
        String sql = String.format("UPDATE %s SET value = ? WHERE code = ? and scada_time = ?", table);
        jdbcTemplate.update(sql, value, code, time);
    }


}
